searchData={"items":[{"type":"module","title":"Ecto.Adapters.FoundationDB","doc":"([Hex.pm](https://hex.pm/packages/ecto_foundationdb) | [GitHub](https://github.com/foundationdb-beam/ecto_foundationdb))\n\nAdapter module for [FoundationDB](https://www.foundationdb.org).\n\nIt uses [erlfdb](https://github.com/foundationdb-beam/erlfdb) for communicating with the database,\nand implements an Ecto-compatible Layer on top of FDB's key-value store.","ref":"Ecto.Adapters.FoundationDB.html"},{"type":"module","title":"Installation - Ecto.Adapters.FoundationDB","doc":"Install the latest stable release of FoundationDB from the\n[official FoundationDB Releases](https://github.com/apple/foundationdb/releases).\n\nThe `foundationdb-server` package is required on any system that will be running\na FoundationDB server instance. For example, it's common to\nrun the `foundationdb-server` on your development machine and on managed\ninstances running a FoundationDB cluster, but not for your stateless Elixir\napplication server in production.\n\n`foundationdb-clients` is always required.\n\nInclude `:ecto_foundationdb` in your list of dependencies in `mix.exs`:\n\n```elixir\ndefp deps do\n  [\n    {:ecto_foundationdb, \"~> 0.4\"}\n  ]\nend\n```","ref":"Ecto.Adapters.FoundationDB.html#module-installation"},{"type":"module","title":"Usage - Ecto.Adapters.FoundationDB","doc":"Define this module in your `MyApp` application:\n\n```elixir\ndefmodule MyApp.Repo do\n  use Ecto.Repo, otp_app: :my_app, adapter: Ecto.Adapters.FoundationDB\n\n  use EctoFoundationDB.Migrator\n  def migrations(), do: []\nend\n```\n\nEdit your `config.exs` file to include the following:\n\n```elixir\nconfig :my_app, MyApp.Repo,\n  cluster_file: \"/etc/foundationdb/fdb.cluster\"\n```\n\nIn your app's supervisor, you can start the repo as follows:\n\n```elixir\nchildren = [\n  ...\n  MyApp.Repo,\n  ...\n]\n\nSupervisor.start_link(...)\n```","ref":"Ecto.Adapters.FoundationDB.html#module-usage"},{"type":"module","title":"Tenants - Ecto.Adapters.FoundationDB","doc":"EctoFoundationDB requires your application to use multitenancy via `EctoFoundationDB.Tenant`.\nWe strongly encourage your application to decide on a multitenancy strategy that makes sense\nfor your problem domain (or simply use `\"default\"` to get started).\n\nOnce you have opened a tenant, you'll use the Ecto `:prefix` option to provide the tenant\nto each Repo operation.","ref":"Ecto.Adapters.FoundationDB.html#module-tenants"},{"type":"module","title":"Creating a schema - Ecto.Adapters.FoundationDB","doc":"This is a standard `Ecto.Schema` that will be used in the examples of this documentation.\n\n```elixir\ndefmodule User do\n  use Ecto.Schema\n  @primary_key {:id, :binary_id, autogenerate: true}\n  schema \"users\" do\n    field(:name, :string)\n    field(:department, :string)\n    field(:balance, :integer)\n    timestamps()\n  end\n  # ...\nend\n```","ref":"Ecto.Adapters.FoundationDB.html#module-creating-a-schema"},{"type":"module","title":"Opening a tenant - Ecto.Adapters.FoundationDB","doc":"See `EctoFoundationDB.Tenant` for more details.\n\n```elixir\nalias EctoFoundationDB.Tenant\n\ntenant = Tenant.open!(MyApp.Repo, \"some-org\")\n```","ref":"Ecto.Adapters.FoundationDB.html#module-opening-a-tenant"},{"type":"module","title":"Inserting a new struct - Ecto.Adapters.FoundationDB","doc":"The Tenant is stored on the struct as metadata, so that when passing structs\nto Repo functions, the `:prefix` option is not required.\n\n```elixir\nalice = %User{name: \"Alice\", department: \"Engineering\"}\n       |> FoundationDB.usetenant(tenant)\n\nalice = MyApp.Repo.insert!(alice)\n```","ref":"Ecto.Adapters.FoundationDB.html#module-inserting-a-new-struct"},{"type":"module","title":"Querying for a struct using the primary key - Ecto.Adapters.FoundationDB","doc":"The `:prefix` option is required. The struct returned will have the tenant\nin the metadata, like above.\n\n```elixir\nMyApp.Repo.get!(User, alice.id, prefix: tenant)\n```","ref":"Ecto.Adapters.FoundationDB.html#module-querying-for-a-struct-using-the-primary-key"},{"type":"module","title":"Transactions - Ecto.Adapters.FoundationDB","doc":"Multiple calls to Repo functions can be grouped inside a single transaction, which is\n[ACID](https://en.wikipedia.org/wiki/ACID) and [globally serializable](https://en.wikipedia.org/wiki/Global_serializability).\n\nHere's a simple example. Transactions can be arbitrarily complex, but there is a [limit to the size and runtime](#module-foundationdb).\n\n```elixir\nMyApp.Repo.transaction(fn ->\n  alice = MyApp.Repo.get!(User, alice.id)\n  if alice.balance > 0 do\n    MyApp.Repo.update(User.changeset(alice, %{balance: alice.balance - 1}))\n  else\n    raise \"Overdraft\"\n  end\nend, prefix: tenant)\n```","ref":"Ecto.Adapters.FoundationDB.html#module-transactions"},{"type":"module","title":"Indexes and Queries - Ecto.Adapters.FoundationDB","doc":"The implication of the [FoundationDB Layer Concept](https://apple.github.io/foundationdb/layer-concept.html)\nis that the manner in which data can be stored and accessed is the responsibility of the client\napplication. To achieve general purpose data storage and access patterns commonly desired\nin web applications, EctoFoundationDB provides support for indexes out of the box, and\nalso allows your application to define custom indexes.\n\nWhen you define a migration that calls `create index(User, [:department])`, a `Default` index is\ncreated on the `:department` field. The following two indexes are equivalent:\n\n```elixir\n# ...\ncreate index(User, [:department])\n# is equivalent to\ncreate index(User, [:department], options: [indexer: EctoFoundationDB.Indexer.Default]))]\n# ...\n```\n\nTimestamp indexes and multi-field indexes are supported:\n\n```elixir\ncreate index(User, [:inserted_at])\ncreate index(User, [:birthyear, :department, :birthdate])\n```\n\nIn fact, the index value can be any Elixir term. All types support Equal queries (of the form `x.field == \"value\"`).\nHowever, certain Ecto Schema types support Between queries (of the form `x.field >= 10 and x.field < 20`).\nSee the [Data Types](#module-data-types) section for the list.\n\nSee the [Migrations](#module-migrations) section for further details about managing indexes in your application.","ref":"Ecto.Adapters.FoundationDB.html#module-indexes-and-queries"},{"type":"module","title":"Index Query Examples - Ecto.Adapters.FoundationDB","doc":"Retrieve all users in the Engineering department (with an Equal constraint):\n\n```elixir\nquery = from(u in User, where: u.department == ^\"Engineering\")\nMyApp.Repo.all(query, prefix: tenant)\n```\n\nRetrieve all users with a birthyear from 1992 to 1994 (with a Between constraint):\n\n```elixir\nquery = from(u in User,\n  where: u.birthyear >= ^1992 and u.birthyear <= ^1994\n)\nMyApp.Repo.all(query, prefix: tenant)\n```\n\nRetrieve all Engineers born in August 1992:\n\n```elixir\nquery = from(u in User,\n  where: u.birthyear == ^1992 and\n         u.department == ^\"Engineering\" and\n         u.birthdate >= ^~D[1992-08-01] and u.birthdate < ^~D[1992-09-01]\n)\nMyApp.Repo.all(query, prefix: tenant)\n```","ref":"Ecto.Adapters.FoundationDB.html#module-index-query-examples"},{"type":"module","title":"Index ordering matters - Ecto.Adapters.FoundationDB","doc":"When you create an index using multiple fields, the FDB key that stores the index will be extended with\nall the values *in the order of your defined index fields*. Because FoundationDB stores the keys themselves\nin a well-defined order, the order of the fields in your index determines the Between queries you can perform.\n\n> **_RULE:_**  For any Query, there can be 0 or 1 Between clauses, and if one exists for a field, then there must not\nexist an Equal clause for ay field that follows, according to the list of fields in the index. Moreover, the Between\nfield must be of a [Data Type](#module-data-types) supported for Between queries.\n\nAbove, we demonstrated a nontrivial query with 2 Equal constraints and 1 Between constraint. Compare that with an invalid\nquery below, which attempts a Between constraint on the `:birthyear` field followed by an Equal constraint on the `:department`\nfield.\n\n```elixir\n# Invalid query!\nfrom(u in User,\n  where: u.birthyear >= ^1992 and u.birthyear < ^1995 and\n         u.department == ^\"Engineering\"\n)\n```\n\n**Why is this query unsupported by EctoFDB?** With the index defined as-is, EctoFDB is unable to extract a minimal\ndataset from the database in one operation. Instead of over-extracting data and then applying a filter, we've chosen\nto reject the query. We hope this encourages correct index management and removes worry about execution time of\nindividual Repo function calls.\n\n**Yeah that's cool, but how do I run the query I want to run?** You have 2 options.\n\n1. Purposefully over-extract the data using Stream, and do the filtering in Elixir:\n\n```elixir\nfrom(u in User, where: u.birthyear >= ^1992 and u.birthyear < 1995 and)\n|> MyApp.Repo.stream(prefix: tenant)\n|> Stream.filter(&((&1).department == \"Engineering\"))\n|> Enum.to_list()\n```\n\nor\n\n2. Create the necessary index. For our example above, the new index might look like:\n\n```elixir\ncreate index(User, [:department, :birthyear])\n```\n\nWe encourage you to test execution times of your expected workload before deciding to create an index.","ref":"Ecto.Adapters.FoundationDB.html#module-index-ordering-matters"},{"type":"module","title":"Custom Indexes - Ecto.Adapters.FoundationDB","doc":"As data is inserted, updated, deleted, and queried, the Indexer callbacks (via behaviour `EctoFoundationDB.Indexer`)\nare executed. Your Indexer module does all the work necessary to maintain the index within the transaction\nthat the data itself is being manipulated. Many different types of indexes are possible. You are in control!\n\nFor an example, please see the [MaxValue implementation](https://github.com/foundationdb-beam/ecto_foundationdb/blob/main/lib/ecto_foundationdb/indexer/max_value.ex)\nwhich is used internally by EctoFoundationDB to keep track of the maximum value of a field in a Schema.","ref":"Ecto.Adapters.FoundationDB.html#module-custom-indexes"},{"type":"module","title":"Transactions - Ecto.Adapters.FoundationDB","doc":"The Repo functions will automatically use FoundationDB transactions behind the scenes. EctoFoundationDB also\nexposes the Ecto.Repo transaction function, allowing you to group operations together with transactional isolation:\n\n```elixir\nMyApp.Repo.transaction(fn ->\n    # Ecto work\n  end, prefix: tenant)\n```\n\nPlease read the [FoundationDB Developer Guide on transactions](https://apple.github.io/foundationdb/developer-guide.html#transaction-basics)\nfor more information about how to develop with transactions.\n\nIt's important to remember that even though the database gives ACID guarantees about the\nkeys updated in a transaction, the Elixir function passed into `Repo.transaction/1`\nmay be executed more than once when any conflicts\nare encountered. For this reason, your function must not have any side effects other than\nthe database operations. For example, do not publish any messages to other processes\n(such as `Phoenix.PubSub`) from within the transaction.\n\n```elixir\n# Do not do this!\nMyApp.Repo.transaction(fn ->\n  MyApp.Repo.insert(%User{name: \"Alice\"})\n  ...\n  Phoenix.PubSub.broadcast(\"my_topic\", \"new_user\") # Not safe! :(\nend, prefix: tenant)\n```\n\n```elixir\n# Instead, do this:\nMyApp.Repo.transaction(fn ->\n  MyApp.Repo.insert(%User{name: \"Alice\"})\n  ...\nend, prefix: tenant)\nPhoenix.PubSub.broadcast(\"my_topic\", \"new_user\") # Safe :)\n```\n\nIf you're looking for PubSub-like functionality at the transaction-level, please see\n[Watches](#module-watches).","ref":"Ecto.Adapters.FoundationDB.html#module-transactions"},{"type":"module","title":"Migrations - Ecto.Adapters.FoundationDB","doc":"At first glance, EctoFoundationDB migrations may look similar to that of `:ecto_sql`,\nbut the actual execution of migrations and how your app must be configured are very\ndifferent, so please read this section in full.\n\nIf your app uses indexes on any of your schemas, you must include a module that implements\nthe `EctoFoundationDB.Migrator` behaviour. By default, this is assumed to be your Repo\nmodule, as we have demonstrated above.\n\nMigrations are only used for index management. There are no tables to add, drop, or modify.\nWhen you add a field to your Schema, any read requests on old objects will return `nil` in\nthat field.\n\nAs tenants are opened during your application runtime, migrations will be executed\nautomatically. This distributes the migration across a potentially long period of time,\nas migrations will not be executed unless the tenant is opened. Specifically, a call to\n`Tenant.open/2` will block until the migration for that tenant completes. Other tenants\nwill not be affected.\n\nThe Migrator is a module in your application runtime that provides the full list of\nordered migrations. These are the migrations that will be executed when a tenant is opened.\nIf you leave out a migration from the list, it will not be applied.\n\nThe result of `migrations/0` will change over time to include new migrations, like so:\n\n```elixir\ndefmodule MyApp.Repo do\n  # ...\n  def migrations() do\n    [\n      {0, MyApp.IndexUserByDepartment}\n      # At a later date, we my add a new index with a corresponding addition\n      # tho the migrations list:\n      #    {1, MyApp.IndexUserByRole}\n    ]\n  end\nend\n```\n\nEach migration is contained in a separate module, much like EctoSQL's. However, the operations\nto be carried out **must be returned as a list.** For example, the creation an index\nmay look like this:\n\n```elixir\ndefmodule MyApp.IndexUserByDepartment do\n  use EctoFoundationDB.Migration\n  def change() do\n    [\n      create index(User, [:department])\n    ]\n  end\nend\n```\n\nYour migrator and `MyApp. ` modules are part of your codebase. They must be\nincluded in your application release.\n\nMigrations can be completed in full at any time with a call to\n`EctoFoundationDB.Migrator.up_all/1`. Depending on how many tenants you have in\nyour database, and the size of the data for each tenant, this operation might take a long\ntime and make heavy use of system resources. It is safe to interrupt this operation, as\nmigrations are performed transactionally. However, if you interrupt a migration, the next\nattempt for that tenant may have a brief delay (~5 sec); the migrator must ensure that the\nprevious migration has indeed stopped executing.","ref":"Ecto.Adapters.FoundationDB.html#module-migrations"},{"type":"module","title":"Data Types - Ecto.Adapters.FoundationDB","doc":"EctoFoundationDB stores your struct's data using `:erlang.term_to_binary/1`, and\nretrieves it with `:erlang.binary_to_term/1`. As such, there is no data type\nconversion between Elixir types and Database Types. Any term you put in your\nstruct will be stored and later retrieved.\n\nData types *are* used by EctoFoundationDB for the creation and querying of indexes.\nCertain Ecto types are encoded into the FDB key, which allows you to formulate\nBetween queries on indexes of these types:\n\n - `:id`\n - `:binary_id`\n - `:integer`\n - `:float`\n - `:boolean`\n - `:string`\n - `:binary`\n - `:date`\n - `:time`\n - `:time_usec`\n - `:naive_datetime`\n - `:naive_datetime_usec`\n - `:utc_datetime`\n - `:utc_datetime_usec`\n\nSee [Indexes and Queries](#module-indexes-and-queries) for more information.","ref":"Ecto.Adapters.FoundationDB.html#module-data-types"},{"type":"module","title":"Watches - Ecto.Adapters.FoundationDB","doc":"[FoundationDB Watches](https://apple.github.io/foundationdb/developer-guide.html#watches) are\nsimilar to Triggers in an RDBMS. Registering a watch on a particular key provides a guarantee[*](https://github.com/apple/foundationdb/wiki/An-Overview-how-Watches-Work)\nthat when that key in the database is changed, the client application is notified with a\npush-style notification. This registration is initiated inside a transaction, and the caller is\nprovided a `EctoFoundationDB.Future` object that is used to receive the push notification at a later time.\nThe notification is delivered directly to the Elixir process that requested it.\n\nThe EctoFDB Repo module allows you to register a watch on any struct that you have written\nto the database. The FDB watch is registered on the key that represents the Primary Write (the key-value\nthat stores all the data of your struct). When that struct changes in the database via any means,\nthe Future is resolved and the caller has the opportunity to refresh its internal state.","ref":"Ecto.Adapters.FoundationDB.html#module-watches"},{"type":"module","title":"EctoFDB Repo functions for watches - Ecto.Adapters.FoundationDB","doc":"EctoFDB adds the following functions to your Repo module. They are each discussed in detail\nbelow.\n\n- `Repo.watch/1`/`Repo.watch/2`: Registers a watch\n- `Repo.assign_ready/2`/`Repo.assign_ready/3`: Handles Future resolution via a push message\n\n### `Repo.watch/2`\n\nWith arguments `struct` and `options`, registers\na watch on the given struct and returns a Future. This Future resolves to\na value when the FDB key is set or cleared. Until then, the Future remains unresolved.\nThe Future from a watch is resolvable outside of an FDB transaction.\n\nIn addition to other standard options, a `:label` option will prime the Future for use by `assign_ready/3`.\nIt's recommended that you provide a label and use `assign_ready/3`, but not required.\n\n### `Repo.assign_ready/3`\n\nWith arguments `futures`, `ready_refs`, and\n`options`, returns a tuple containing: 1) a Keyword of structs with keys according to the `:label`\nprovided to `watch/2`, and 2) an updated list of futures that are still unresolved.\n\nIn order for the caller to acquire the `ready_refs` arguments, you must receive messages from the\nprocess mailbox of the form `{ref, :ready}`. See the example below for detail. If your process chooses\nto buffer a list `ready_refs` longer than 1, then `assign_ready/3` uses pipelining to retrieve the\nresulting structs. Each watched struct must have been registered with a unique `:label`.\n\nThe `options` are provided to `Repo.async_get/3` internally. In addition to standard options, a\n`watch?: true` option will re-register a new watch on the struct so that your calling\nprocess's event loop can always be aware of changes.","ref":"Ecto.Adapters.FoundationDB.html#module-ectofdb-repo-functions-for-watches"},{"type":"module","title":"Example - Ecto.Adapters.FoundationDB","doc":"The power of watches is best illustrated in the context of a stateful Elixir process that is\ninterested in tracking the up-to-date value for a particular schema. This could be a GenServer, LiveView,\nor some other stateful process.\n\n```elixir\ndefmodule MyAliceView do\n\n  # ...\n\n  def init(tenant) do\n    {alice, future} = MyRepo.transaction(\n                        fn ->\n                          alice = MyRepo.get_by(User, name: \"Alice\")\n                          {alice, MyRepo.watch(alice, label: :alice)}\n                        end, prefix: tenant)\n    assigns = [alice: alice]\n    {:ok, %{assigns: assigns, futures: [future]}}\n  end\n\n  # ...\n\n  def handle_info({ref, :ready}, state=%{assigns: assigns, futures: futures}) when is_reference(ref) do\n    {new_assigns, new_futures} = MyRepo.assign_ready(futures, [ref], watch?: true)\n    {:noreply, %{\n          state |\n            assigns: Map.merge(assigns, Enum.into(new_assigns, %{})),\n            futures: new_futures\n          }}\n  end\nend\n```\n\nBy handling the Future message resolution event `{ref, :ready}` in the `handle_info/2`, we never need to\ncall `Repo.await/1`. Instead, our process can do other work and react to the `:ready` event immediately\nwhen it's delivered.\n\nAnd `Repo.assign_ready/3` provides a conventional way to retrieve the result and register a new watch\nin one go, so that updates to Alice are not missed.","ref":"Ecto.Adapters.FoundationDB.html#module-example"},{"type":"module","title":"Upserts - Ecto.Adapters.FoundationDB","doc":"The FoundationDB Adapter supports the following upsert approaches.","ref":"Ecto.Adapters.FoundationDB.html#module-upserts"},{"type":"module","title":"Upserts with `on_conflict` - Ecto.Adapters.FoundationDB","doc":"The following choices for `on_conflict` are supported, and they work in the manner\ndescribed by the official Ecto documentation.\n\n- `:raise`: The default.\n- `:nothing`\n- `:replace_all`\n- `{:replace_all_except, fields}`\n- `{:replace, fields}`","ref":"Ecto.Adapters.FoundationDB.html#module-upserts-with-on_conflict"},{"type":"module","title":"Upserts with `conflict_target` - Ecto.Adapters.FoundationDB","doc":"If you provide `conflict_target: []`, the FoundationDB Adapter will make no effort to\ninspect the existence of objects in the database before insert. When an existing\nobject is blindly overwritten, your indexes may become inconsistent, resulting in undefined\nbehavior. You should only consider its use if one of the following is true:\n\n- (a) your schema has no indexes\n- (b) you know apriori that any indexed fields are unchanged at the time of upsert, or\n- (c) you know apirori that the primary key you're inserting doesn't already exist.\n\nFor these use cases, this option may speed up the loading of initial data at scale.","ref":"Ecto.Adapters.FoundationDB.html#module-upserts-with-conflict_target"},{"type":"module","title":"Pipelining - Ecto.Adapters.FoundationDB","doc":"Pipelining is the technique of sending multiple queries to a database at the same time, and\nthen receiving the results at a later time. In doing so, you can often avoid waiting\nfor multiple network round trips. EctoFDB uses a `async`/`await` syntax to\nimplement pipelining.","ref":"Ecto.Adapters.FoundationDB.html#module-pipelining"},{"type":"module","title":"EctoFDB Repo functions for pipelining - Ecto.Adapters.FoundationDB","doc":"EctoFDB adds the following functions to your Repo module. You'll notice that there are `async_*`\nversions of all the Queryable functions on the standard `Ecto.Repo`. These have the same arguments\nas their counterparts, and they each return a single `EctoFoundationDB.Future`.\n\n- `Repo.async_get/2` / `Repo.async_get/3`\n- `Repo.async_get!/2` / `Repo.async_get!/3`\n- `Repo.async_get_by/2` / `Repo.async_get_by/3`\n- `Repo.async_get_by!/2` / `Repo.async_get_by!/3`\n- `Repo.async_one/1` / `Repo.async_one/2`\n- `Repo.async_one!/1` / `Repo.async_one!/2`\n- `Repo.async_all/1` / `Repo.async_all/2`\n- `Repo.await/1`: The results of the above functions are provided to `Repo.await/1` to\n  resolve the Futures. It accepts a single Future or a list of Futures.","ref":"Ecto.Adapters.FoundationDB.html#module-ectofdb-repo-functions-for-pipelining"},{"type":"module","title":"Example - Ecto.Adapters.FoundationDB","doc":"Consider this example. Our transaction below has 2 network round trips in serial.\nThe \"Alice\" User is retrieved, and then the \"Bob\" user is retrieved. Finally, the\nlist is constructed and the transaction ends.\n\n```elixir\nresult = MyApp.Repo.transaction(fn ->\n  [\n    MyApp.Repo.get_by(User, name: \"Alice\"),\n    MyApp.Repo.get_by(User, name: \"Bob\")\n  ]\nend, prefix: tenant)\n\n[alice, bob] = result\n```\n\nWith pipelining, we can issue 2 `async_get_by` queries without waiting for their result,\nand then only when we need the data, we `await` the result. Notice that the list of futures\nis constructed, and then we `await` on them.\n\n```elixir\nresult = MyApp.Repo.transaction(fn ->\n  futures = [\n    MyApp.Repo.async_get_by(User, name: \"Alice\"),\n    MyApp.Repo.async_get_by(User, name: \"Bob\")\n  ]\n\n  MyApp.Repo.await(futures)\nend, prefix: tenant)\n\n[alice, bob] = result\n```\n\nConceptually, only 1 network round-trip is required, so this will be faster than\nwaiting on each in sequence.","ref":"Ecto.Adapters.FoundationDB.html#module-example"},{"type":"module","title":"Pipelining on range queries (interleaving waits) - Ecto.Adapters.FoundationDB","doc":"The functions `Repo.async_all/1` and `Repo.async_all/2` work a little bit different on the await. When the DB must\nreturn a list of results, if that list ends up being larger than `:erlfdb`'s configured `:target_bytes`,\nthen the result set is split into multiple round-trips to the database. If your application awaits multiple\nsuch queries, then `:erlfdb`'s \"interleaving waits\" feature is used.\n\nConsider this example, where we're retrieving all Users and all Posts in the same transaction.\n\n```elixir\nMyApp.Repo.transaction(fn ->\n  futures = [\n    MyApp.Repo.async_all(User),\n    MyApp.Repo.async_all(Post)\n  ]\n\n  MyApp.Repo.await(futures)\nend, prefix: tenant)\n```\n\nEach query individually executed would be multiple round-trips to the database as pages of results are retrieved.\n\n```mermaid\nsequenceDiagram\n    Client->>Server: get_range(User)\n    Server-->>Client: {100 users, continuation_token}\n    Client->>Server: get_range(User, continuation_token)\n    Server-->>Client: {10 users, done}\n```\n\nWith `:erlfdb`'s interleaving waits, we send as many get_range requests as possible with pipelining\nuntil all are exhausted.\n\n```mermaid\nsequenceDiagram\n    Client->>Server: get_range(User)\n    Client->>Server: get_range(Post)\n    Server-->>Client: {100 users, continuation_token}\n    Server-->>Client: {100 posts, continuation_token}\n    Client->>Server: get_range(User, continuation_token)\n    Client->>Server: get_range(Post, continuation_token)\n    Server-->>Client: {10 users, done}\n    Server-->>Client: {100 posts, continuation_token}\n    Client->>Server: get_range(Post, continuation_token)\n    Server-->>Client: {2 posts, done}\n```\n\nThis happens automatically when you include the list of futures to your `Repo.await/1`.","ref":"Ecto.Adapters.FoundationDB.html#module-pipelining-on-range-queries-interleaving-waits"},{"type":"module","title":"Standard Options - Ecto.Adapters.FoundationDB","doc":"* `:cluster_file` - The path to the fdb.cluster file. The default is\n    `\"/etc/foundationdb/fdb.cluster\"`.\n  * `:migrator` - A module that implements the `EctoFoundationDB.Migrator`\n    behaviour. Required when using any indexes. Defaults to `nil`. When `nil`,\n    your Repo module is assumed to be the Migrator.","ref":"Ecto.Adapters.FoundationDB.html#module-standard-options"},{"type":"module","title":"Advanced Options - Ecto.Adapters.FoundationDB","doc":"* `:storage_id` - All tenants created by this adapter are prefixed with\n    this string. This allows multiple configurations to operate on the\n    same FoundationDB cluster independently. Defaults to\n    `\"Ecto.Adapters.FoundationDB\"`.\n  * `:open_db` - 1-arity function accepting the Repo module. Used for opening a reference to the\n     FoundationDB cluster. Defaults to `&EctoFoundationDB.Database.open/1`. When\n     using `EctoFoundationDB.Sandbox`, you should set this option to `Sandbox.open_db/1`.\n  * `:migration_step` - The maximum number of keys to process in a single transaction\n    when running migrations. Defaults to `1000`. If you use a number that is\n    too large, the FDB transactions run by the Migrator will fail.\n  * `:idx_cache` - When set to `:enabled`, the Ecto ets cache is used to store the\n    available indexes per tenant. This speeds up all database operations.\n    Defaults to `:enabled`.\n  * `:max_single_value_size` - Number of Bytes above which an encoded struct will\n    be split into multiple FDB key-value pairs upon insert and update. Defaults to `100_000`.\n    Any value beyond `100_000` is unsupported by FoundationDB. You probably do not\n    want to change this value. Also see `:max_value_size`.\n  * `:max_value_size` - Number of Bytes above which an encoded struct will be\n    rejected by EctoFDB upon insert and update. Defaults to `:infinity`, which\n    means that a single encoded struct of any size can be written to any number of\n    keys. Please remember that FDB's transaction size limit of 10MB still applies.\n    The maximum number of key-value pairs for a given encoded struct, not counting\n    indexes, is computed by `1 + ceil(:max_value_size / :max_single_value_size)`.","ref":"Ecto.Adapters.FoundationDB.html#module-advanced-options"},{"type":"module","title":"Optimizing for throughput and latency - Ecto.Adapters.FoundationDB","doc":"Pipelining operations inside your transactions can offer significant speed-ups, and you should\nalso consider parallelizing multiple transactions. Both techniques are useful in different\ncircumstances.\n\n- Pipelining: The transaction can contain arbitrary logic and remain ACID compliant. If you make\n  smart use of pipelining, the latency of execution of that logic is kept to a minimum.\n- Parallel transactions: Each transaction is internally ACID compliant, but any 2 given\n  transactions may conflict if they're accessing the same keys. When transactions do not conflict,\n  throughput goes up.\n\nIf you're attempting to get the maximum throughput on data loading, you'll probably want to\nmake use of both!\n\nRead more at [FDB Developer Guide | Loading Data](https://apple.github.io/foundationdb/developer-guide.html#loading-data).","ref":"Ecto.Adapters.FoundationDB.html#module-optimizing-for-throughput-and-latency"},{"type":"module","title":"Limitations and Caveats - Ecto.Adapters.FoundationDB","doc":"","ref":"Ecto.Adapters.FoundationDB.html#module-limitations-and-caveats"},{"type":"module","title":"FoundationDB - Ecto.Adapters.FoundationDB","doc":"Please read [FoundationDB Known Limitations](https://apple.github.io/foundationdb/known-limitations.html).","ref":"Ecto.Adapters.FoundationDB.html#module-foundationdb"},{"type":"module","title":"Layer - Ecto.Adapters.FoundationDB","doc":"EctoFoundationDB implements a specific Layer on the FoundationDB\nkey-value store. This Layer is intended to be generally useful, but you\nmay not find it suitable for your workloads. The Layer is documented in\ndetail at `EctoFoundationDB.Layer`. This project does not support\nplugging in other Layers.","ref":"Ecto.Adapters.FoundationDB.html#module-layer"},{"type":"module","title":"Tenants - Ecto.Adapters.FoundationDB","doc":"* The use of a tenant implies an ownership relationship between the tenant and the data.\n   It's up to you how you use this relationship in your application.\n\n * Failure to specify a tenant will result in a raised exception at runtime.","ref":"Ecto.Adapters.FoundationDB.html#module-tenants"},{"type":"module","title":"Migrations - Ecto.Adapters.FoundationDB","doc":"The following are not supported:\n\n* Renaming a table\n* Renaming a field\n* Dropping an index\n* Moving down in migration bersions (rollback)\n\nFinally, the Ecto Mix tasks are known to be unsupported:\n\n```elixir\n# These commands are not supported. Do not use them with :ecto_foundationdb!\n#    mix ecto.create\n#    mix ecto.drop\n#    mix ecto.migrate\n#    mix ecto.gen.migration\n#    mix ecto.rollback\n#    mix ecto.migrations\n```","ref":"Ecto.Adapters.FoundationDB.html#module-migrations"},{"type":"module","title":"Key and Value Size - Ecto.Adapters.FoundationDB","doc":"[FoundationDB has limits on key and value size](https://apple.github.io/foundationdb/known-limitations.html#large-keys-and-values)\nPlease be aware of these limitations as you develop. EctoFoundationDB doesn't make\nany attempt to protect you from these errors.","ref":"Ecto.Adapters.FoundationDB.html#module-key-and-value-size"},{"type":"module","title":"Ecto Queries - Ecto.Adapters.FoundationDB","doc":"You'll find that most queries outside of those detailed in the documentation fail with\nan `EctoFoundationDB.Exception.Unsupported` exception. The FoundationDB Layer concept\nprecludes complex queries from being executed within the database. Therefore, it only makes sense\nto implement a limited set of query types -- specifically those that do have optimized database\nquery semantics. All other filtering, aggregation, and grouping must be done by your Elixir code.\n\nIn other words, you'll be writing less SQL, and more Elixir.\n\nRefer to the integration tests for a collection of queries that is known to work.","ref":"Ecto.Adapters.FoundationDB.html#module-ecto-queries"},{"type":"module","title":"Other Ecto Features - Ecto.Adapters.FoundationDB","doc":"As you test the boundaries of EctoFoundationDB, you'll find that many other Ecto features just plain\ndon't work with this adapter. Ecto has a lot of features, and this adapter is in the early stage.\nFor any given unsupported feature, the underlying reason will be one of the following.\n\n1. It makes sense for EctoFoundationDB, but it just hasn't been implemented yet.\n2. The fundamental differences between FDB and SQL backends mean that there is\n   no practical reason to implement it in EctoFoundationDB.","ref":"Ecto.Adapters.FoundationDB.html#module-other-ecto-features"},{"type":"function","title":"Ecto.Adapters.FoundationDB.db/1","doc":"","ref":"Ecto.Adapters.FoundationDB.html#db/1"},{"type":"function","title":"Ecto.Adapters.FoundationDB.transactional/2","doc":"Executes the given function in a transaction on the database.\n\nIf you provide an arity-0 function, your function will be executed in\na newly spawned process. This is to ensure that EctoFoundationDB can\nsafely manage the process dictionary.\n\nPlease be aware of the\n[limitations that FoundationDB](https://apple.github.io/foundationdb/developer-guide.html#transaction-basics)\nimposes on transactions.\n\nFor example, a transaction must complete\n[within 5 seconds](https://apple.github.io/foundationdb/developer-guide.html#long-running-transactions).","ref":"Ecto.Adapters.FoundationDB.html#transactional/2"},{"type":"function","title":"Ecto.Adapters.FoundationDB.usetenant/2","doc":"","ref":"Ecto.Adapters.FoundationDB.html#usetenant/2"},{"type":"module","title":"EctoFoundationDB.Database","doc":"See `Ecto.Adapters.FoundationDB`","ref":"EctoFoundationDB.Database.html"},{"type":"function","title":"EctoFoundationDB.Database.open/1","doc":"","ref":"EctoFoundationDB.Database.html#open/1"},{"type":"type","title":"EctoFoundationDB.Database.t/0","doc":"","ref":"EctoFoundationDB.Database.html#t:t/0"},{"type":"module","title":"EctoFoundationDB.Future","doc":"Opaque struct that represents an unresolved result set from some FoundationDB\nquery operation.\n\nIf you have received a Future from an EctoFDB Repo API, please consult the documentation\nfor that API for dealing with the Future. The functions here are intended to be for\ninternal use only.","ref":"EctoFoundationDB.Future.html"},{"type":"function","title":"EctoFoundationDB.Future.apply/2","doc":"","ref":"EctoFoundationDB.Future.html#apply/2"},{"type":"function","title":"EctoFoundationDB.Future.await_all/1","doc":"","ref":"EctoFoundationDB.Future.html#await_all/1"},{"type":"function","title":"EctoFoundationDB.Future.await_stream/1","doc":"","ref":"EctoFoundationDB.Future.html#await_stream/1"},{"type":"function","title":"EctoFoundationDB.Future.before_transactional/1","doc":"","ref":"EctoFoundationDB.Future.html#before_transactional/1"},{"type":"function","title":"EctoFoundationDB.Future.find_ready/2","doc":"","ref":"EctoFoundationDB.Future.html#find_ready/2"},{"type":"function","title":"EctoFoundationDB.Future.leaving_transactional/1","doc":"","ref":"EctoFoundationDB.Future.html#leaving_transactional/1"},{"type":"function","title":"EctoFoundationDB.Future.new/1","doc":"","ref":"EctoFoundationDB.Future.html#new/1"},{"type":"function","title":"EctoFoundationDB.Future.new_watch/3","doc":"","ref":"EctoFoundationDB.Future.html#new_watch/3"},{"type":"function","title":"EctoFoundationDB.Future.ref/1","doc":"","ref":"EctoFoundationDB.Future.html#ref/1"},{"type":"function","title":"EctoFoundationDB.Future.result/1","doc":"","ref":"EctoFoundationDB.Future.html#result/1"},{"type":"function","title":"EctoFoundationDB.Future.schema/1","doc":"","ref":"EctoFoundationDB.Future.html#schema/1"},{"type":"function","title":"EctoFoundationDB.Future.set/4","doc":"","ref":"EctoFoundationDB.Future.html#set/4"},{"type":"function","title":"EctoFoundationDB.Future.token/0","doc":"","ref":"EctoFoundationDB.Future.html#token/0"},{"type":"module","title":"EctoFoundationDB.Index","doc":"See `Ecto.Adapters.FoundationDB`","ref":"EctoFoundationDB.Index.html"},{"type":"type","title":"EctoFoundationDB.Index.t/0","doc":"","ref":"EctoFoundationDB.Index.html#t:t/0"},{"type":"behaviour","title":"EctoFoundationDB.Indexer","doc":"Implement this behaviour to create a custom index.","ref":"EctoFoundationDB.Indexer.html"},{"type":"callback","title":"EctoFoundationDB.Indexer.clear/5","doc":"","ref":"EctoFoundationDB.Indexer.html#c:clear/5"},{"type":"function","title":"EctoFoundationDB.Indexer.clear/6","doc":"","ref":"EctoFoundationDB.Indexer.html#clear/6"},{"type":"function","title":"EctoFoundationDB.Indexer.create/6","doc":"","ref":"EctoFoundationDB.Indexer.html#create/6"},{"type":"callback","title":"EctoFoundationDB.Indexer.create/6","doc":"","ref":"EctoFoundationDB.Indexer.html#c:create/6"},{"type":"function","title":"EctoFoundationDB.Indexer.create_range/2","doc":"","ref":"EctoFoundationDB.Indexer.html#create_range/2"},{"type":"callback","title":"EctoFoundationDB.Indexer.create_range/2","doc":"","ref":"EctoFoundationDB.Indexer.html#c:create_range/2"},{"type":"function","title":"EctoFoundationDB.Indexer.range/3","doc":"","ref":"EctoFoundationDB.Indexer.html#range/3"},{"type":"callback","title":"EctoFoundationDB.Indexer.range/3","doc":"","ref":"EctoFoundationDB.Indexer.html#c:range/3"},{"type":"callback","title":"EctoFoundationDB.Indexer.set/5","doc":"","ref":"EctoFoundationDB.Indexer.html#c:set/5"},{"type":"function","title":"EctoFoundationDB.Indexer.set/6","doc":"","ref":"EctoFoundationDB.Indexer.html#set/6"},{"type":"function","title":"EctoFoundationDB.Indexer.unpack/3","doc":"","ref":"EctoFoundationDB.Indexer.html#unpack/3"},{"type":"callback","title":"EctoFoundationDB.Indexer.unpack/3","doc":"","ref":"EctoFoundationDB.Indexer.html#c:unpack/3"},{"type":"callback","title":"EctoFoundationDB.Indexer.update/6","doc":"","ref":"EctoFoundationDB.Indexer.html#c:update/6"},{"type":"function","title":"EctoFoundationDB.Indexer.update/7","doc":"","ref":"EctoFoundationDB.Indexer.html#update/7"},{"type":"module","title":"EctoFoundationDB.Layer","doc":"The Ecto FoundationDB Adapter implements a [Layer](https://apple.github.io/foundationdb/layer-concept.html)\non the underlying key-value store provided by FoundationDB. Via this layer, some\ncommon data access patterns are achieved. For those familiar with relational databases\nsuch as Postgres, these patterns will be familiar. However, there are many differences (for example SQL is\nnot supported), so this document seeks to describe the capabilities of the Ecto FoundationDB Layer in detail.","ref":"EctoFoundationDB.Layer.html"},{"type":"module","title":"Keyspace Design - EctoFoundationDB.Layer","doc":"All keys used by `:ecto_foundationdb` are encoded with [FoundationDB's Tuple Encoding](https://github.com/apple/foundationdb/blob/main/design/tuple.md).\n\nWith the default `:tenant_backend` of `EctoFoundationDB.Tenant.DirectoryTenant`, each tuple written by EctoFDB is\nprefixed with a short binary string as allocated by `:erlfdb_directory. This tenant prefixing is the most critical\nelement of the keyspace because it's the mechanism that guarantees tenants cannot cross their boundaries. **The rest of\nthis documentation is written assuming you're using the DirectoryTenant`.**\n\nThe first element of the tuple after the tenant prefix is a string prefix that is intended to keep the\n`:ecto_foundationdb` keyspace separate from other keys in the FoundationDB cluster.\n\nYour Schema data and Default indexes are stored with \"\\xFD\".\n\nThe data associated with schema migrations is stored with \"\\xFE\".\n\nThe rest of the tenant's keyspace is open for use by you, the application developer. For example, it is safe to write:\n\n```elixir\ndb = FoundationDB.open(MyApp.Repo)\ntenant = Tenant.open(MyApp.Repo, \"some-org\")\n\n# Multitenancy-Safe. The key is properly packed\n:erlfdb.set(db, Tenant.pack(tenant, {\"hello\"}), \"world\")\n\n# Multitenancy-Unsafe. The key is not packed into the tenant's keyspace\n:erlfdb.set(db, :erlfdb_tuple.pack({\"hello\"}), \"world\")\n:erlfdb.set(db, \"hello\", \"world\")\n```\n\nA value (the binary stored at each key) is either\n  * some other keys (in the case of Default indexes) or\n  * Erlang term data encoded with `:erlang.term_to_binary/1`","ref":"EctoFoundationDB.Layer.html#module-keyspace-design"},{"type":"module","title":"Primary Write and Read - EctoFoundationDB.Layer","doc":"Your Ecto Schema has a primary key field, which is usually a string or an integer. This primary\nkey uniquely identifies an object of your schema within the tenant in which it lives.\n\n```elixir\ndefmodule EctoFoundationDB.Schemas.User do\n  use Ecto.Schema\n  schema \"users\" do\n    field(:name, :string)\n    field(:department, :string)\n    timestamps()\n  end\nend\n```\n\nIn this example, a User has an `:id` and a `:name`. Also remember that the User is defined within\na tenant which provides a scope under which the User lives. For example, a typical tenant\nwould be the organization the User belongs to. Since the User is in this tenant, we do not need\nto provide an identifier for this organization on the User object itself.\n\n\"Primary Write\" refers to the insertion of the User struct into the FoundationDB key-value store\nunder a single key that uniquely identifies the User. This key includes the `:id` value.\n\nYour struct data is stored as a `Keyword` encoded with `:erlang.term_to_binary/1`.\n\nNote: The Primary Write can be skipped by providing the `write_primary: false` option on the `@schema_context`.\nSee below for more.","ref":"EctoFoundationDB.Layer.html#module-primary-write-and-read"},{"type":"module","title":"Default Indexes - EctoFoundationDB.Layer","doc":"When a Default index is created via a migration, the Ecto FoundationDB Adapter writes a set of\nkeys and values to facilitate lookups based on the indexed field.","ref":"EctoFoundationDB.Layer.html#module-default-indexes"},{"type":"module","title":"Advanced Options: `write_primary: false`, `mapped?: false` - EctoFoundationDB.Layer","doc":"If you choose to use `write_primary: false` on your schema, this skips the Primary Write. The consequence of this are as follows:\n\n 1. You'll want to make sure your index is created with an option `mapped?: false`. This ensures that the\n    struct data is written to the index keyspace.\n 2. Your index queries will now have performance characteristics similar to a primary key query. That is, you'll\n    be able to retrieve the struct data with a single `erlfdb:get_range/3`.\n 3. The data can **only** be managed by providing a query on the index. You will not be able to access the data\n    via the primary key.\n 4. If `write_primary: false`, then only one index can be created.","ref":"EctoFoundationDB.Layer.html#module-advanced-options-write_primary-false-mapped-false"},{"type":"behaviour","title":"EctoFoundationDB.Migrator","doc":"Implement this behaviour to define migrations for `Ecto.Adapters.FoundationDB`","ref":"EctoFoundationDB.Migrator.html"},{"type":"callback","title":"EctoFoundationDB.Migrator.migrations/0","doc":"","ref":"EctoFoundationDB.Migrator.html#c:migrations/0"},{"type":"function","title":"EctoFoundationDB.Migrator.up/3","doc":"","ref":"EctoFoundationDB.Migrator.html#up/3"},{"type":"function","title":"EctoFoundationDB.Migrator.up_all/2","doc":"","ref":"EctoFoundationDB.Migrator.html#up_all/2"},{"type":"module","title":"EctoFoundationDB.Options","doc":"See `Ecto.Adapters.FoundationDB`","ref":"EctoFoundationDB.Options.html"},{"type":"function","title":"EctoFoundationDB.Options.get/2","doc":"","ref":"EctoFoundationDB.Options.html#get/2"},{"type":"type","title":"EctoFoundationDB.Options.option/0","doc":"","ref":"EctoFoundationDB.Options.html#t:option/0"},{"type":"type","title":"EctoFoundationDB.Options.t/0","doc":"","ref":"EctoFoundationDB.Options.html#t:t/0"},{"type":"module","title":"EctoFoundationDB.QueryPlan","doc":"See `Ecto.Adapters.FoundationDB`","ref":"EctoFoundationDB.QueryPlan.html"},{"type":"function","title":"EctoFoundationDB.QueryPlan.get/7","doc":"","ref":"EctoFoundationDB.QueryPlan.html#get/7"},{"type":"function","title":"EctoFoundationDB.QueryPlan.get_op/3","doc":"","ref":"EctoFoundationDB.QueryPlan.html#get_op/3"},{"type":"function","title":"EctoFoundationDB.QueryPlan.get_pinned_param/2","doc":"","ref":"EctoFoundationDB.QueryPlan.html#get_pinned_param/2"},{"type":"function","title":"EctoFoundationDB.QueryPlan.walk_ast/4","doc":"","ref":"EctoFoundationDB.QueryPlan.html#walk_ast/4"},{"type":"type","title":"EctoFoundationDB.QueryPlan.t/0","doc":"","ref":"EctoFoundationDB.QueryPlan.html#t:t/0"},{"type":"module","title":"EctoFoundationDB.Sandbox","doc":"A module for managing a sandbox FoundationDB cluster. This allows a developer to create\na space under which it should be safe to write tests.\n\nWhen using this module, it creates a directory in the project root called `.erlfdb`. It\nis safe to delete this directory when you no longer need it (e.g. after test execution)\n\nSee [Testing with EctoFoundationDB](testing.html) for more.","ref":"EctoFoundationDB.Sandbox.html"},{"type":"function","title":"EctoFoundationDB.Sandbox.checkin/2","doc":"","ref":"EctoFoundationDB.Sandbox.html#checkin/2"},{"type":"function","title":"EctoFoundationDB.Sandbox.checkout/3","doc":"","ref":"EctoFoundationDB.Sandbox.html#checkout/3"},{"type":"function","title":"EctoFoundationDB.Sandbox.open_db/1","doc":"","ref":"EctoFoundationDB.Sandbox.html#open_db/1"},{"type":"module","title":"EctoFoundationDB.Tenant","doc":"This module allows the application to create, open, and delete\ntenants within the FoundationDB database. All transactions require a tenant,\nso any application that uses the Ecto FoundationDB Adapter must use this module.\n\nEctoFoundationDB supports 2 different backends for the Tenant implementation:\n\n1. `EctoFoundationDB.Tenant.DirectoryTenant`: The default. Tenants are managed via\n   the `:erlfdb_directory` layer. Each tenant is assigned a byte prefix. EctoFDB\n   puts that prefix as the first element of every tuple key. If your application wishes\n   to write keys directly to the database, it must also respect this prefix.\n2. `EctoFoundationDB.Tenant.ManagedTenant`: Uses the experimental [FoundationDB Tenants](https://apple.github.io/foundationdb/tenants.html).\n   This requires `tenant_mode=required_experimental` or `tenant_mode=optional_experimental`\n   to be enabled on your FDB system. EctoFDB does not need to prefix keys because it happens\n   automatically at the FDB transaction level. If your application wishes to write keys\n   directory to the database, there is no prefix to worry about. **Use caution:** FDB Tenants\n   are not yet tested with the same rigor as the rest of FDB.\n\nIf you wish to use the experimental ManagedTenant, add this option to your Repo config:\n\n```elixir\nconfig :my_app, MyApp.Repo,\n  tenant_backend: EctoFoundationDB.Tenant.ManagedTenant\n```\n\nEctoFDB does not support switching between tenant backends on a database. If you want to switch backend,\nyou must use a new empty database.","ref":"EctoFoundationDB.Tenant.html"},{"type":"function","title":"EctoFoundationDB.Tenant.clear/2","doc":"Clear all data for the given tenant. This cannot be undone.","ref":"EctoFoundationDB.Tenant.html#clear/2"},{"type":"function","title":"EctoFoundationDB.Tenant.clear_delete!/2","doc":"Clears data in a tenant and then deletes it. If the tenant doesn't exist, no-op.","ref":"EctoFoundationDB.Tenant.html#clear_delete!/2"},{"type":"function","title":"EctoFoundationDB.Tenant.create/2","doc":"Create a tenant in the database.","ref":"EctoFoundationDB.Tenant.html#create/2"},{"type":"function","title":"EctoFoundationDB.Tenant.delete/2","doc":"Deletes a tenant from the database permanently. The tenant must\nhave no data.","ref":"EctoFoundationDB.Tenant.html#delete/2"},{"type":"function","title":"EctoFoundationDB.Tenant.exists?/2","doc":"Returns true if the tenant already exists in the database.","ref":"EctoFoundationDB.Tenant.html#exists?/2"},{"type":"function","title":"EctoFoundationDB.Tenant.list/1","doc":"List all tenants in the database. Could be expensive.","ref":"EctoFoundationDB.Tenant.html#list/1"},{"type":"function","title":"EctoFoundationDB.Tenant.open/3","doc":"Open a tenant with a repo. With the result returned by this function, the caller can\ndo database operations on the tenant's portion of the key-value store.\n\nThe tenant must already exist.\n\nWhen opening tenants with a repo, all migrations are automatically performed. This\ncan cause open/2 to take a significant amount of time. Tenants can be kept open\nindefinitely, with any number of database transactions issued upon them.","ref":"EctoFoundationDB.Tenant.html#open/3"},{"type":"function","title":"EctoFoundationDB.Tenant.open!/3","doc":"Open a tenant. With the result returned by this function, the caller can\ndo database operations on the tenant's portion of the key-value store.\n\nIf the tenant does not exist, it is created.\n\nWhen opening tenants with a repo, all migrations are automatically performed. This\ncan cause open/2 to take a significant amount of time. Tenants can be kept open\nindefinitely, with any number of database transactions issued upon them.","ref":"EctoFoundationDB.Tenant.html#open!/3"},{"type":"function","title":"EctoFoundationDB.Tenant.open_empty!/3","doc":"Helper function to ensure the given tenant exists and then clear\nit of all data, and finally return an open handle. Useful in test code,\nbut in production, this would be dangerous.","ref":"EctoFoundationDB.Tenant.html#open_empty!/3"},{"type":"function","title":"EctoFoundationDB.Tenant.pack/2","doc":"Packs an Elixir tuple into an FDB-encoded Tuple.","ref":"EctoFoundationDB.Tenant.html#pack/2"},{"type":"function","title":"Keyspace Design - EctoFoundationDB.Tenant.pack/2","doc":"We always pack into a **non-prefixed tuple key**. In other words, EctoFDB\nDirectoryTenant keys use the subspace prefix as the first tuple element\ninstead of binary key prefix. As such, our keys are not compliant with\nother Directory Layer implementations. However, we've made this choice so that\nwe can continue to use GetMappedRange functionality. (Indeed, the GetMappedRange mapper spec\nhas reserved syntax for stripping out non-tuple prefixes, but it's not yet implemented.)\nNote: ManagedTenant keys do not use directories/subspaces. The underlying binary\nprefix that a ManagedTenant uses internally still allows use of GetMappedRange.\n\nManagedTenants and GetMappedRange are both experimental features. The safest choice is\nto use neither, but that would forfeit the GetMappedRange optimization. We choose\nto accept the risk of GetMappedRange, which can easily be replaced with a different\nclient implementation, and suggest against using ManagedTenant because it puts\nyour data at risk with its unsupported[[0](https://github.com/apple/foundationdb/issues/11292)][[1](https://github.com/apple/foundationdb/issues/11382)]\nkeyspace.","ref":"EctoFoundationDB.Tenant.html#pack/2-keyspace-design"},{"type":"function","title":"EctoFoundationDB.Tenant.primary_codec/2","doc":"","ref":"EctoFoundationDB.Tenant.html#primary_codec/2"},{"type":"function","title":"EctoFoundationDB.Tenant.primary_mapper/1","doc":"","ref":"EctoFoundationDB.Tenant.html#primary_mapper/1"},{"type":"function","title":"EctoFoundationDB.Tenant.range/2","doc":"","ref":"EctoFoundationDB.Tenant.html#range/2"},{"type":"function","title":"EctoFoundationDB.Tenant.txobj/1","doc":"","ref":"EctoFoundationDB.Tenant.html#txobj/1"},{"type":"function","title":"EctoFoundationDB.Tenant.unpack/2","doc":"","ref":"EctoFoundationDB.Tenant.html#unpack/2"},{"type":"type","title":"EctoFoundationDB.Tenant.id/0","doc":"","ref":"EctoFoundationDB.Tenant.html#t:id/0"},{"type":"type","title":"EctoFoundationDB.Tenant.prefix/0","doc":"","ref":"EctoFoundationDB.Tenant.html#t:prefix/0"},{"type":"type","title":"EctoFoundationDB.Tenant.t/0","doc":"","ref":"EctoFoundationDB.Tenant.html#t:t/0"},{"type":"module","title":"EctoFoundationDB.Tenant.DirectoryTenant","doc":"The default backend for EctoFDB multitenancy.\n\nIt uses the Directory Layer defined by `:erlfdb_directory` to\npartition tenant keyspaces from each other. There is no guarantee\nprovided by FDB at the transaction-level, so care must be taken to\nalways pack keys appropriately. Standard use of Ecto.Repo functions\nis guaranteed to respect tenant boundaries.","ref":"EctoFoundationDB.Tenant.DirectoryTenant.html"},{"type":"type","title":"EctoFoundationDB.Tenant.DirectoryTenant.t/0","doc":"","ref":"EctoFoundationDB.Tenant.DirectoryTenant.html#t:t/0"},{"type":"module","title":"EctoFoundationDB.Tenant.ManagedTenant","doc":"An experimental backend for EctoFDB multitenancy. It uses FDB's Tenants\nto guarantee that a given transaction cannot access keys from another tenant.\n\nTo use ManagedTenant, your database must be configured with\n\n```shell\nfdbcli --exec 'configure tenant_mode=required_experimental'\n# or\nfdbcli --exec 'configure tenant_mode=optional_experimental'\n```\n\nFDB's tenants are still an experimental feature, so EctoFDB's ManagedTenant must\nalso be considered experimental. For now, we recommend using `EctoFoundationDB.Tenant.DirectoryTenant`,\nwhich is the default.","ref":"EctoFoundationDB.Tenant.ManagedTenant.html"},{"type":"type","title":"EctoFoundationDB.Tenant.ManagedTenant.t/0","doc":"","ref":"EctoFoundationDB.Tenant.ManagedTenant.html#t:t/0"},{"type":"exception","title":"EctoFoundationDB.Exception.IncorrectTenancy","doc":"This exception is raised when there is a conflict in the tenant used.","ref":"EctoFoundationDB.Exception.IncorrectTenancy.html"},{"type":"exception","title":"EctoFoundationDB.Exception.Unsupported","doc":"This exception is raised when the application uses an Ecto feature that is not\nsupported by the FoundationDB Adapter.","ref":"EctoFoundationDB.Exception.Unsupported.html"},{"type":"extras","title":"Changelog for v0.x","doc":"# Changelog for v0.x","ref":"changelog.html"},{"type":"extras","title":"v0.4 (2024-01-16) - Changelog for v0.x","doc":"","ref":"changelog.html#v0-4-2024-01-16"},{"type":"extras","title":"Enhancements - Changelog for v0.x","doc":"* [Large Structs](Ecto.Adapters.FoundationDB.html#module-advanced-options): If your struct encodes to a size larger than 100,000 Bytes, it will now be split across several FDB key-values automatically.\n  Previously, EctoFDB did not attempt to detect this and allowed FoundationDB to throw error code 2103: \"value_too_large - Value length exceeds limit\".\n  See documentation for configuration options.\n* `EctoFoundationDB.Sandbox` now uses `m:erlfdb_sandbox`. Sandbox directory name is now `.erlfdb_sandbox`. Directories named `.erlfdb` should be removed.","ref":"changelog.html#enhancements"},{"type":"extras","title":"Bug fixes - Changelog for v0.x","doc":"* Upgrade erlfdb to v0.2.2","ref":"changelog.html#bug-fixes"},{"type":"extras","title":"v0.3.1 (2024-10-24) - Changelog for v0.x","doc":"","ref":"changelog.html#v0-3-1-2024-10-24"},{"type":"extras","title":"Bug fixes - Changelog for v0.x","doc":"* Fixed consistency issue with index updates. Previously, the old index key was still queryable.\n* Fixed write amplification issue when updating struct's non-indexed fields.","ref":"changelog.html#bug-fixes"},{"type":"extras","title":"New Documentation - Changelog for v0.x","doc":"* Added `fdb_api_counting_text.exs` which tests and documents the `:erlfdb` operations that our Layer is expected to make.","ref":"changelog.html#new-documentation"},{"type":"extras","title":"v0.3 (2024-10-20) - Changelog for v0.x","doc":"### \\*\\* Major breaking changes \\*\\*\n\nDatabases that have been created using a prior version of EctoFoundationDB will be broken on\nEctoFDB v0.3 and above. Please start a new database with EctoFDB v0.3. If you currently have\na database on v0.2 or earlier, please [submit an issue](https://github.com/foundationdb-beam/ecto_foundationdb/issues)\nto discuss the upgrade path.","ref":"changelog.html#v0-3-2024-10-20"},{"type":"extras","title":"Enhancements - Changelog for v0.x","doc":"* [Watches](Ecto.Adapters.FoundationDB.html#module-watches): Support for FDB Watches, which is like a database-driven PubSub on an Ecto struct.\n* [Directory Tenants](EctoFoundationDB.Tenant.html): A new default backend for Multitenancy that is production-ready. Managed tenants have been moved to \"Experimental\" status.\n* `@schema_context usetenant: true` is no longer required.\n* The `:open_db` option now defines a 1-arity function that accepts the Repo module.","ref":"changelog.html#enhancements"},{"type":"extras","title":"New Documentation - Changelog for v0.x","doc":"* [Livebook | Watches in LiveView](watches.livemd)","ref":"changelog.html#new-documentation"},{"type":"extras","title":"v0.2.1 (2024-09-23) - Changelog for v0.x","doc":"","ref":"changelog.html#v0-2-1-2024-09-23"},{"type":"extras","title":"Bug fixes - Changelog for v0.x","doc":"* Upgrade erlfdb to 0.2.1, allowing Livebook Desktop to discover fdbcli location in /usr/local/bin","ref":"changelog.html#bug-fixes"},{"type":"extras","title":"v0.2.0 (2024-09-21) - Changelog for v0.x","doc":"","ref":"changelog.html#v0-2-0-2024-09-21"},{"type":"extras","title":"Bug fixes - Changelog for v0.x","doc":"","ref":"changelog.html#bug-fixes"},{"type":"extras","title":"Enhancements - Changelog for v0.x","doc":"* [Upserts](Ecto.Adapters.FoundationDB.html#module-upserts): Support for Ecto options `:on_conflict` and `:conflict_target`\n  * [Pipelining](Ecto.Adapters.FoundationDB.html#module-pipelining): New Repo functions for async/await within a transaction.","ref":"changelog.html#enhancements"},{"type":"extras","title":"New Documentation - Changelog for v0.x","doc":"* [TESTING.md](testing.html): Document to describe how to set up a Sandbox\n  * [CHANGELOG.md](changelog.html): This file!","ref":"changelog.html#new-documentation"},{"type":"extras","title":"Testing with EctoFoundationDB","doc":"# Testing with EctoFoundationDB\n\nSetting up your app to use a FoundationDB Sandbox is very easy!\n\nFirst, in `config/test.exs`:\n\n```elixir\nconfig :my_app, MyApp.Repo,\n  open_db: &EctoFoundationDB.Sandbox.open_db/1\n```\n\nNext, set up an ExUnit Case that will provide you with a new Tenant to use\nin your tests. This step is optional, but recommended: it will make writing\neach test frictionless.\n\n```elixir\ndefmodule MyApp.TenantCase do\n  @moduledoc false\n  use ExUnit.CaseTemplate\n\n  alias EctoFoundationDB.Sandbox\n\n  setup do\n    tenant_id = Ecto.UUID.autogenerate()\n\n    tenant = Sandbox.checkout(MyApp.Repo, tenant_id, [])\n\n    on_exit(fn ->\n      Sandbox.checkin(MyApp.Repo, tenant_id)\n    end)\n\n    {:ok, [tenant_id: tenant_id, tenant: tenant]}\n  end\nend\n```\n\nNow, you can use your TenantCase to do any FDB operation in a test. Because\nwe're using tenants with randomized names, you don't have to worry about key conflicts.\n\n```elixir\ndefmodule MyAppHelloFDBTest do\n  use MyApp.TenantCase\n\n  alias Ecto.Adapters.FoundationDB\n\n  test \"hello\", context do\n    tenant = context[:tenant]\n\n    # An :erlfdb hello world, delete this for your tests\n    assert :not_found ==\n             FoundationDB.transactional(\n               tenant,\n               fn tx ->\n                 tx\n                 |> :erlfdb.get(\"hello world\")\n                 |> :erlfdb.wait()\n               end\n             )\n\n    # Example:\n    #   MyApp.Repo.insert!(%MyApp.Hello{message: \"world\"}, prefix: tenant)\n  end\nend\n```","ref":"testing.html"},{"type":"extras","title":"Livebook | Getting Started","doc":"# Livebook | Getting Started\n\n```elixir\nMix.install([\n  {:ecto_foundationdb, \"~> 0.3\"}\n])\n```","ref":"guide.html"},{"type":"extras","title":"Setup - Livebook | Getting Started","doc":"Hello! This guide simulates what your experience might be when developing an application with EctoFoundationDB. Specifically, it focuses on the mechanism that EctoFoundationDB uses to create and manage indexes.\n\nIt assumes the reader is familiar with general Ecto features.\n\nBefore we get started, a couple of important points about executing these commands on your system.\n\n> If you received an error on the `Mix.install` setup, please make sure you have both `foundationdb-server` and `foundationdb-clients` packages installed on your system. Also, ensure that your Livebook PATH environment variable includes the directory containing the `fdbcli` binary.\n\n> This LiveBook expects your system to have a running instance of FoundationDB, and it writes and deletes data from it. If your system's `/etc/foundationdb/fdb.cluster` is pointing to a real database, do not execute these commands!\n\nWith that out of the way, we'll start off with creating your Repo module.\n\n```elixir\ndefmodule MyApp.Repo do\n  use Ecto.Repo, otp_app: :my_app, adapter: Ecto.Adapters.FoundationDB\n\n  use EctoFoundationDB.Migrator\n\n  def migrations() do\n    [\n      # {0, IndexesMigration}\n    ]\n  end\nend\n```\n\nNotice that the line with `IndexesMigration` is commented out. We'll come back to this later.","ref":"guide.html#setup"},{"type":"extras","title":"Developing your app - Livebook | Getting Started","doc":"This next step simulates your app's startup. Normally, you would have a project defining `:my_app` and the Repo would be included in your supervision tree. In this Guide, we're starting the Repo as an isolated resource.\n\n```elixir\n{:ok, _} = Ecto.Adapters.FoundationDB.ensure_all_started(MyApp.Repo.config(), :temporary)\nMyApp.Repo.start_link(log: false)\n```\n\nNext, we define an `Ecto.Schema` for events that are coming from a temperature sensor. This is a pretty standard Schema module.\n\n```elixir\ndefmodule TemperatureEvent do\n  use Ecto.Schema\n\n  @primary_key {:id, :binary_id, autogenerate: true}\n\n  schema \"temperature_events\" do\n    field(:recorded_at, :naive_datetime_usec)\n    field(:kelvin, :float)\n    field(:site, :string)\n    timestamps()\n  end\nend\n```\n\nWe're going to create a module that will help us insert some `TemperatureEvents`.\n\n```elixir\ndefmodule Sensor do\n  alias Ecto.Adapters.FoundationDB\n\n  def record(n, tenant) do\n    MyApp.Repo.transaction(\n      fn ->\n        for _ <- 1..n, do: record(nil)\n      end,\n      prefix: tenant\n    )\n  end\n\n  def record(tenant) do\n    %TemperatureEvent{\n      site: \"surface\",\n      kelvin: 373.15 + :rand.normal(0, 5),\n      recorded_at: NaiveDateTime.utc_now()\n    }\n    |> FoundationDB.usetenant(tenant)\n    |> MyApp.Repo.insert!()\n  end\nend\n```\n\nNow, we create and open a new Tenant to store our `TemperatureEvents`.\n\n```elixir\nalias EctoFoundationDB.Tenant\n\ntenant = Tenant.open!(MyApp.Repo, \"experiment-42c\")\n```\n\nWe're ready to record an event from our temperature sensor! Feel free to Reevaluate this block several times. You'll record 4 new events each time.\n\n```elixir\nfor _ <- 1..4, do: Sensor.record(tenant)\n```\n\nWe can list all the events from the Tenant. This uses a single FoundationDB Transaction.\n\n```elixir\nMyApp.Repo.all(TemperatureEvent, prefix: tenant)\n```\n\nIf there's a large number of events, you can stream them instead of reading them all at once. This uses multiple FoundationDB Transactions.\n\n```elixir\nMyApp.Repo.stream(TemperatureEvent, prefix: tenant)\n|> Enum.to_list()\n|> length()\n```\n\nNext, we'd like to read all events from `\"surface\"`. If you're executing this LiveBook in order, you'll receive an exception on this step.\n\n```elixir\nimport Ecto.Query\n\nquery = from(e in TemperatureEvent, where: e.site == ^\"surface\")\nMyApp.Repo.all(query, prefix: tenant)\n```\n\nDid you get an exception? If so, scroll back up to the `defmodule MyApp.Repo` block in the Setup section, un-comment the line with `IndexesMigration`, and Reevaluate that block. Then come back and continue from here. You don't need to Reevaluate other blocks above this text.\n\n👋\n\nWelcome back! You've instructed the Repo to load a migration next time we open a Tenant. But we still need to define that Migration. The block below defines two indexes.\n\n```elixir\ndefmodule IndexesMigration do\n  use EctoFoundationDB.Migration\n\n  def change() do\n    [\n      create(index(TemperatureEvent, [:site])),\n      create(index(TemperatureEvent, [:recorded_at]))\n    ]\n  end\nend\n```\n\nNow, we re-open the Tenant. **Something very important happens here.**\n\nThis block simulates you restarting your app, and your client reconnecting. We'll just simply call `open!/2` again.\n\n```elixir\ntenant = Tenant.open!(MyApp.Repo, \"experiment-42c\")\n```\n\nGreat! If you made it to this step, then the Migration has executed automatically, and the indexes are ready to be used.","ref":"guide.html#developing-your-app"},{"type":"extras","title":"Querying your data - Livebook | Getting Started","doc":"This next block has the same query as the one that threw an exception earlier. This time, you should retrieve the expected events.\n\n```elixir\nimport Ecto.Query\nquery = from(e in TemperatureEvent, where: e.site == ^\"surface\")\nMyApp.Repo.all(query, prefix: tenant)\n```\n\nWe can also use the timestamp index that we created in a new query.\n\n```elixir\nnow = NaiveDateTime.utc_now()\npast = NaiveDateTime.add(now, -1200, :second)\n\nquery =\n  from(e in TemperatureEvent,\n    where: e.recorded_at >= ^past and e.recorded_at < ^now\n  )\n\nMyApp.Repo.all(query, prefix: tenant)\n```\n\nFinally, just for fun, let's insert 10,000 `TemperatureEvent`s!\n\n```elixir\nnum = 10000\nbatch = 100\n\n{t, :ok} =\n  :timer.tc(fn ->\n    Stream.duplicate(batch, div(num, batch))\n    |> Task.async_stream(\n      Sensor,\n      :record,\n      [tenant],\n      max_concurrency: System.schedulers_online() * 8,\n      ordered: false,\n      timeout: 30000\n    )\n    |> Stream.run()\n  end)\n\nIO.puts(\"Done in #{t / 1000} msec\")\n```","ref":"guide.html#querying-your-data"},{"type":"extras","title":"Cleaning up - Livebook | Getting Started","doc":"And if you'd like to tidy up, you can easily delete all the data.\n\n```elixir\n# Note: destructive!\nMyApp.Repo.delete_all(TemperatureEvent, prefix: tenant)\n```\n\n```elixir\n# Note: destructive!\nTenant.clear_delete!(MyApp.Repo, \"experiment-42c\")\n```","ref":"guide.html#cleaning-up"},{"type":"extras","title":"Livebook | Watches in LiveView","doc":"# Livebook | Watches in LiveView\n\n```elixir\nMix.install([\n  {:phoenix_playground, \"~> 0.1.0\"},\n  {:ecto_foundationdb, \"~> 0.3\"}\n])\n```","ref":"watches.html"},{"type":"extras","title":"Intro - Livebook | Watches in LiveView","doc":"We're going to create a simple LiveView to showcase your favorite quote. We want to make sure our LiveView always has the most up-to-date information about the quote. A common appproach to a problem like this is to use `Phoenix.PubSub`. Instead, we'll use [EctoFoundationDB's Watches](https://hexdocs.pm/ecto_foundationdb/Ecto.Adapters.FoundationDB.html#module-watches) to deliver the messaging.","ref":"watches.html#intro"},{"type":"extras","title":"Setup Ecto - Livebook | Watches in LiveView","doc":"First, we set up Ecto, defining a Quote schema, and starting the Repo. For this to work, you must have `foundationdb-server` running locally. Refer to the [EctoFoundationDB documentation](https://hexdocs.pm/ecto_foundationdb/Ecto.Adapters.FoundationDB.html#module-installation) for installation help.\n\n```elixir\ndefmodule Repo do\n  use Ecto.Repo, otp_app: :my_app, adapter: Ecto.Adapters.FoundationDB\n\n  use EctoFoundationDB.Migrator\n\n  def migrations(), do: []\nend\n\ndefmodule Quote do\n  use Ecto.Schema\n\n  alias Ecto.Changeset\n\n  @primary_key {:id, :binary_id, autogenerate: true}\n\n  schema \"quotes\" do\n    field(:author, :string)\n    field(:content, :string)\n    field(:likes, :integer, default: 0)\n    field(:as_of, :naive_datetime)\n    timestamps()\n  end\n\n  def changeset(quote, params \\\\ %{}) do\n    quote\n    |> Changeset.cast(params, [:author, :content, :likes])\n  end\n\n  def like!(tenant, id) do\n      Repo.transaction(fn ->\n        quote = Repo.get!(Quote, id)\n\n        quote\n        |> changeset(%{likes: quote.likes+1})\n        |> Repo.update!()\n      end, prefix: tenant)\n  end\nend\n\n\n{:ok, _} = Ecto.Adapters.FoundationDB.ensure_all_started(Repo.config(), :temporary)\nRepo.start_link(log: false)\n\nalias EctoFoundationDB.Tenant\n\ntenant = Tenant.open!(Repo, \"experiment-with-watches-in-liveview\")\n\n```\n\nNext, we create a Quote in the DB. We'll focus on how this specific Quote is rendered throughout the rest of the Livebook.\n\n```elixir\nRepo.insert!(\n  %Quote{\n    id: \"my-favorite-quote\",\n    author: \"Philippe Verdoux\",\n    content: \"\"\"\n      Enlightenment leads to benightedness; Science entails nescience.\n      \"\"\",\n    as_of: NaiveDateTime.utc_now(:second)\n  }, prefix: tenant, on_conflict: :replace_all)\n```","ref":"watches.html#setup-ecto"},{"type":"extras","title":"Setup LiveView - Livebook | Watches in LiveView","doc":"We're using Phoenix Playground to create a sample LiveView. Some key takeaways:\n\n1. In `assign_watch!/4`, we're reading the Quote with id `\"my-favorite-quote\"` that we've inserted above. In the same transaction, we're creating a watch with `label: :quote`. In the LiveView assigns, we store the `:quote` and the list of `:futures`. You'll want to use the same label for both the watch and the assigns.\n\n2. `handle_event/3` receives the `\"like\"` event, and uses an FDB transaction to add 1 to the count of likes. Notice we choose not to update the assigns, for demonstration purposes.\n\n3. `handle_info/2` receives the `:ready` message from the watch future. It uses `Repo.assign_ready/3` to update the LiveView assigns map. Whenever `\"my-favorite-quote\"` changes in the DB, this function will be called automatically. Also, we provide `watch?: true` to continue listening for updates.\n\n4. `assign_watch!/4` and `handle_info/2` both contain nothing specific to the Quote schema. This is a general purpose approach that can be re-used for any other schema types.\n\n```elixir\ndefmodule DemoLive do\n  use Phoenix.LiveView\n\n  def mount(_params, _session, socket) do\n\n    tenant = Tenant.open!(Repo, \"experiment-with-watches-in-liveview\")\n\n    {:ok, socket\n      |> assign(tenant: tenant)\n      |> assign_watch!(Quote, :quote, \"my-favorite-quote\")\n    }\n  end\n\n  defp assign_watch!(socket, schema, label, id) do\n    {struct, futures} = Repo.transaction(\n      fn ->\n        struct = Repo.get!(schema, id)\n        future = Repo.watch(struct, label: label)\n        {struct, [future]}\n      end,\n      prefix: socket.assigns.tenant)\n\n    socket\n    |> assign(label, struct)\n    |> assign(futures: futures)\n  end\n\n  def render(assigns) do\n    ~H\"\"\"\n     \n       My Favorite Quote \n       as of <%= @quote.as_of %> UTC \n       \n         <%= @quote.content %> \n          - <%= @quote.author %>  \n\n         \n           Likes:  \n           <%= @quote.likes %> \n           👍 \n         \n       \n     \n\n     \n      h1 { text-align: center; }\n      h3 { text-align: right; }\n      h6 { text-align: right; }\n      p { text-align: center; }\n\n      .content {\n        max-width: 500px;\n        margin: auto;\n      }\n     \n    \"\"\"\n  end\n\n  def handle_event(\"like\", _params, socket) do\n\n    Quote.like!(socket.assigns.tenant, socket.assigns.quote.id)\n\n    {:noreply, socket}\n  end\n\n  def handle_info({ref, :ready}, socket) when is_reference(ref) do\n\n    %{assigns: assigns} = socket\n\n    {new_assigns, futures} =\n      Repo.assign_ready(\n        assigns.futures,\n        [ref],\n        watch?: true,\n        prefix: assigns.tenant\n      )\n\n    {:noreply,\n      socket\n      |> assign(new_assigns)\n      |> assign(futures: futures)\n    }\n  end\nend\n\nPhoenixPlayground.start(live: DemoLive)\n```\n\nWith the block above evaluated, you can now open a web browser to `http://localhost:4000`. We suggest you keep this open side-by-side with this Livebook if possible, so that you can watch the updates in real-time.","ref":"watches.html#setup-liveview"},{"type":"extras","title":"Updating My Favorite Quote - Livebook | Watches in LiveView","doc":"Your favorite quote might change from moment to moment. We'll update the database with your current favorite quote, and because of the FDB watch, the LiveView will always render the most up-to-date content, without any PubSub.\n\n```elixir\ntenant = Tenant.open!(Repo, \"experiment-with-watches-in-liveview\")\n\ndefmodule ChangeFavoriteQuote do\n  @id \"my-favorite-quote\"\n\n  def to(tenant, author, content) do\n    Repo.transaction(fn ->\n      Repo.get(Quote, @id)\n      |> Quote.changeset(%{\n        author: author,\n        content: content,\n        as_of: NaiveDateTime.utc_now(:second),\n        likes: 0\n      })\n      |> Repo.update()\n      end, prefix: tenant)\n  end\nend\n```\n\n```elixir\nChangeFavoriteQuote.to(\n  tenant,\n  \"Duke Leto Atreides\",\n  \"\"\"\n  Once you have explored a fear, it becomes less terrifying. \\\n  Part of courage comes from extending our knowledge.\n  \"\"\")\n```\n\n```elixir\nChangeFavoriteQuote.to(tenant, \"Captain America\", \"I can do this all day.\")\n```","ref":"watches.html#updating-my-favorite-quote"},{"type":"extras","title":"Takeaways - Livebook | Watches in LiveView","doc":"The behavior that our webpage exhibits is standard LiveView real-time updating. In real Phoenix applications, this is commonly done via `Phoenix.PubSub`. There are many advantages to using PubSub, but in this demo we replaced it with a feature from FDB.\n\nThe main takeaway from this Livebook is that there is no dispatcher. Each process of our application can choose to register a watch, and it will be notified directly as needed.","ref":"watches.html#takeaways"}],"content_type":"text/markdown","producer":{"name":"ex_doc","version":[48,46,51,52,46,50]}}